# ERPNext - web based ERP (http://erpnext.com)
# Copyright (C) 2012 Web Notes Technologies Pvt Ltd
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
import webnotes
from webnotes import _

from webnotes.utils import cint, cstr, date_diff, flt, formatdate, getdate, get_url_to_form
from webnotes import msgprint

class LeaveDayBlockedError(Exception): pass
	
from webnotes.model.controller import DocListController
class DocType(DocListController):
	def setup(self):
		if webnotes.conn.exists(self.doc.doctype, self.doc.name):
			self.previous_doc = webnotes.doc(self.doc.doctype, self.doc.name)
		else:
			self.previous_doc = None
		
	def validate(self):
		self.validate_to_date()
		self.validate_balance_leaves()
		self.validate_leave_overlap()
		self.validate_max_days()
		self.validate_block_days()
		
	def on_update(self):
		if (not self.previous_doc and self.doc.leave_approver) or (self.previous_doc and \
				self.doc.status == "Open" and self.previous_doc.leave_approver != self.doc.leave_approver):
			# notify leave approver about creation
			self.notify_leave_approver()
		elif self.previous_doc and \
				self.previous_doc.status == "Open" and self.doc.status == "Rejected":
			# notify employee about rejection
			self.notify_employee(self.doc.status)
	
	def on_submit(self):
		if self.doc.status != "Approved":
			webnotes.msgprint("""Only Leave Applications with status 'Approved' can be Submitted.""",
				raise_exception=True)

		# notify leave applier about approval
		self.notify_employee(self.doc.status)
				
	def on_cancel(self):
		# notify leave applier about cancellation
		self.notify_employee("cancelled")

	def validate_block_days(self):
		from hr.doctype.leave_block_list.leave_block_list import get_applicable_block_dates

		block_dates = get_applicable_block_dates(self.doc.from_date, self.doc.to_date, 
			self.doc.employee, self.doc.company)
			
		if block_dates:
			webnotes.msgprint(_("Following dates are blocked for Leave") + ":")
			for d in block_dates:
				webnotes.msgprint(formatdate(d.block_date) + ": " + d.reason)
				
			if self.doc.docstatus == 1:
				raise LeaveDayBlockedError
			
	def get_holidays(self):
		tot_hol = webnotes.conn.sql("""select count(*) from `tabHoliday` h1, `tabHoliday List` h2, `tabEmployee` e1 
			where e1.name = %s and h1.parent = h2.name and e1.holiday_list = h2.name 
			and h1.holiday_date between %s and %s""", (self.doc.employee, self.doc.from_date, self.doc.to_date))
		if not tot_hol:
			tot_hol = webnotes.conn.sql("""select count(*) from `tabHoliday` h1, `tabHoliday List` h2 
				where h1.parent = h2.name and h1.holiday_date between %s and %s
				and ifnull(h2.is_default,0) = 1 and h2.fiscal_year = %s""",
				(self.doc.from_date, self.doc.to_date, self.doc.fiscal_year))
		return tot_hol and flt(tot_hol[0][0]) or 0

	def get_total_leave_days(self):
		"""Calculates total leave days based on input and holidays"""
		ret = {'total_leave_days' : 0.5}
		if not self.doc.half_day:
			tot_days = date_diff(self.doc.to_date, self.doc.from_date) + 1
			holidays = self.get_holidays()
			ret = {
				'total_leave_days' : flt(tot_days)-flt(holidays)
			}
		return ret

	def validate_to_date(self):
		if self.doc.from_date and self.doc.to_date and \
				(getdate(self.doc.to_date) < getdate(self.doc.from_date)):
			msgprint("To date cannot be before from date")
			raise Exception
			
	def validate_balance_leaves(self):
		if self.doc.from_date and self.doc.to_date and not is_lwp(self.doc.leave_type):
			self.doc.leave_balance = get_leave_balance(self.doc.employee,
				self.doc.leave_type, self.doc.fiscal_year)["leave_balance"]
			self.doc.total_leave_days = self.get_total_leave_days()["total_leave_days"]
			
			if self.doc.total_leave_days == 0:
				msgprint(_("Hurray! The day(s) on which you are applying for leave \
					coincide with holiday(s). You need not apply for leave."),
					raise_exception=1)
			
			if self.doc.leave_balance - self.doc.total_leave_days < 0:
				msgprint("There is not enough leave balance for Leave Type: %s" % \
					(self.doc.leave_type,), raise_exception=1)

	def validate_leave_overlap(self):
		for d in webnotes.conn.sql("""select name, leave_type, posting_date, 
			from_date, to_date 
			from `tabLeave Application` 
			where 
			(from_date <= %(to_date)s and to_date >= %(from_date)s)
			and employee = %(employee)s
			and docstatus < 2
			and status in ("Open", "Approved")
			and name != %(name)s""", self.doc.fields, as_dict = 1):
 
			msgprint("Employee : %s has already applied for %s between %s and %s on %s. Please refer Leave Application : <a href=\"#Form/Leave Application/%s\">%s</a>" % (self.doc.employee, cstr(d['leave_type']), formatdate(d['from_date']), formatdate(d['to_date']), formatdate(d['posting_date']), d['name'], d['name']), raise_exception = 1)

	def validate_max_days(self):
		max_days = webnotes.conn.sql("select max_days_allowed from `tabLeave Type` where name = '%s'" %(self.doc.leave_type))
		max_days = max_days and flt(max_days[0][0]) or 0
		if max_days and self.doc.total_leave_days > max_days:
			msgprint("Sorry ! You cannot apply for %s for more than %s days" % (self.doc.leave_type, max_days))
			raise Exception
			
	def notify_employee(self, status):
		employee = webnotes.doc("Employee", self.doc.employee)
		if not employee.user_id:
			return
			
		def _get_message(url=False):
			if url:
				name = get_url_to_form(self.doc.doctype, self.doc.name)
			else:
				name = self.doc.name
				
			return (_("Leave Application") + ": %s - %s") % (name, _(status))
		
		self.notify({
			# for post in messages
			"message": _get_message(url=True),
			"message_to": employee.user_id,
			
			"subject": _get_message(),
		})
		
	def notify_leave_approver(self):
		employee = webnotes.doc("Employee", self.doc.employee)
		
		def _get_message(url=False):
			name = self.doc.name
			employee_name = cstr(employee.employee_name)
			if url:
				name = get_url_to_form(self.doc.doctype, self.doc.name)
				employee_name = get_url_to_form("Employee", self.doc.employee, label=employee_name)
			
			return (_("New Leave Application") + ": %s - " + _("Employee") + ": %s") % (name, employee_name)
		
		self.notify({
			# for post in messages
			"message": _get_message(url=True),
			"message_to": self.doc.leave_approver,
			
			# for email
			"subject": _get_message()
		})
		
	def notify(self, args):
		args = webnotes._dict(args)
		from utilities.page.messages.messages import post
		post({"txt": args.message, "contact": args.message_to, "subject": args.subject,
			"notify": True})

@webnotes.whitelist()
def get_leave_balance(employee, leave_type, fiscal_year):	
	leave_all = webnotes.conn.sql("""select total_leaves_allocated 
		from `tabLeave Allocation` where employee = %s and leave_type = %s
		and fiscal_year = %s and docstatus = 1""", (employee, 
			leave_type, fiscal_year))
	
	leave_all = leave_all and flt(leave_all[0][0]) or 0
	
	leave_app = webnotes.conn.sql("""select SUM(total_leave_days) 
		from `tabLeave Application` 
		where employee = %s and leave_type = %s and fiscal_year = %s
		and docstatus = 1""", (employee, leave_type, fiscal_year))
	leave_app = leave_app and flt(leave_app[0][0]) or 0
	
	ret = {'leave_balance': leave_all - leave_app}
	return ret

@webnotes.whitelist()
def get_approver_list():
	roles = [r[0] for r in webnotes.conn.sql("""select distinct parent from `tabUserRole`
		where role='Leave Approver'""")]
	if not roles:
		webnotes.msgprint("No Leave Approvers. Please assign 'Leave Approver' Role to atleast one user.")
		
	return roles

def is_lwp(leave_type):
	lwp = webnotes.conn.sql("select is_lwp from `tabLeave Type` where name = %s", leave_type)
	return lwp and cint(lwp[0][0]) or 0
	
@webnotes.whitelist()
def get_events(start, end):
	events = []
	employee = webnotes.conn.get_default("employee", webnotes.session.user)
	company = webnotes.conn.get_default("company", webnotes.session.user)

	add_department_leaves(events, start, end, employee, company)
	add_block_dates(events, start, end, employee, company)
	return events
	
def add_department_leaves(events, start, end, employee, company):
	department = webnotes.conn.get_value("Employee", employee, "department")
	
	if not department:
		return
	
	# department leaves
	department_employees = webnotes.conn.sql_list("select name from tabEmployee where department=%s", 
		department)
	
	for d in webnotes.conn.sql("""select name, from_date, to_date, employee_name, half_day, 
		status, employee
		from `tabLeave Application` where
		(from_date between %s and %s or to_date between %s and %s)
		and docstatus < 2
		and status!="Rejected"
		and employee in ('%s')""" % ("%s", "%s", "%s", "%s", "', '".join(department_employees)), 
			(start, end, start, end), as_dict=True):
			events.append({
				"name": d.name,
				"doctype": "Leave Application",
				"from_date": d.from_date,
				"to_date": d.to_date,
				"status": d.status,
				"title": _("Leave by") + " " +  cstr(d.employee_name) + \
					(d.half_day and _(" (Half Day)") or "")
			})
	

def add_block_dates(events, start, end, employee, company):
	# block days
	from hr.doctype.leave_block_list.leave_block_list import get_applicable_block_dates

	cnt = 0
	block_dates = get_applicable_block_dates(start, end, employee, company, all_lists=True)

	for block_date in block_dates:
		events.append({
			"doctype": "Leave Block List Date",
			"from_date": block_date.block_date,
			"title": _("Leave Blocked") + ": " + block_date.reason,
			"name": "_" + str(cnt),
		})
		cnt+=1
	
